@@ -1,10 +1,3 @@
-# Copyright (c) 2014, cPanel, Inc.
-# All rights reserved.
-# http://cpanel.net/
-#
-# This is free software; you can redistribute it and/or modify it under the same
-# terms as Perl itself.  See the LICENSE file for further details.
-
 package Filesys::POSIX;
 
 use strict;
@@ -20,11 +13,14 @@ use Filesys::POSIX::IO       ();
 use Filesys::POSIX::Mount    ();
 use Filesys::POSIX::Userland ();
 
-use Filesys::POSIX::Error qw(throw);
+use Carp qw/confess/;
 
-use Carp qw(confess);
+our $AUTOLOAD;
 
-our $VERSION = '0.9.19_0001';
+BEGIN {
+    use Exporter ();
+    our $VERSION = '0.10.0';
+}
 
 =head1 NAME
 
@@ -85,10 +81,8 @@ again to the VFS, where the options will be stored for later retrieval.
 
 =head1 ERROR HANDLING
 
-Errors are emitted in the form of exceptions thrown by
-L<C<Carp::confess()>|Carp/confess>, with full stack traces.  Where possible,
-L<C<$!>|perlvar/$!> is set with an appropriate error code from L<Errno>, and a
-stringified L<C<$!>|perlvar/$!> is thrown.
+Errors are emitted in the form of exceptions thrown by C<Carp::confess()>, with
+full stack traces.
 
 =cut
 
@@ -113,6 +107,56 @@ sub new {
     }, $class;
 }
 
+sub AUTOLOAD {
+    my ( $self, @args ) = @_;
+    my $method = $AUTOLOAD;
+    $method =~ s/^([a-z0-9_]+::)*//i;
+
+    my $module = $self->{'methods'}->{$method};
+
+    return if $method eq 'DESTROY';
+    confess( "No module imported for method '" . __PACKAGE__ . "::$method()" ) unless $module;
+
+    no strict 'refs';
+
+    return *{"$module\::$method"}->( $self, @args );
+}
+
+=head1 IMPORTING MODULES FOR ADDITIONAL FUNCTIONALITY
+
+=over
+
+=item C<$fs-E<gt>import_module($module)>
+
+Import functionality from the module specified into the namespace of the
+current filesystem object instance.  The module to be imported should be
+specified in the usual form of a Perl package name.  Only the methods returned
+by its EXPORT() function will be imported.
+
+See the L</"EXTENSION MODULES"> section below for a listing of modules that
+Filesys::POSIX provides.
+
+=back
+
+=cut
+
+sub import_module {
+    my ( $self, $module ) = @_;
+
+    eval "use $module";
+    confess $@ if $@;
+
+    no strict 'refs';
+
+    foreach ( *{"$module\::EXPORT"}->() ) {
+        if ( my $imported = $self->{'methods'}->{$_} ) {
+            confess("Module $imported already imported method $_") unless $module eq $imported;
+        }
+
+        $self->{'methods'}->{$_} = $module;
+    }
+}
+
 =head1 SYSTEM CALLS
 
 =over
@@ -148,7 +192,7 @@ sub _find_inode {
         #
         # We've encountered an absolute path.  Start from the beginning.
         #
-        unless (length $item) {
+        unless ($item) {
             $dir = $self->{'root'};
             next;
         }
@@ -173,22 +217,16 @@ sub _find_inode {
         }
         elsif ( $item eq '..' ) {
             my $vnode = $self->{'vfs'}->vnode($dir);
-            $inode =
-                $vnode->{'parent'}
-              ? $vnode->{'parent'}
-              : $self->{'vfs'}->vnode( $directory->get('..') );
+            $inode = $vnode->{'parent'} ? $vnode->{'parent'} : $self->{'vfs'}->vnode( $directory->get('..') );
         }
         else {
             $inode = $self->{'vfs'}->vnode( $directory->get($item) );
         }
 
-        $! = 0;
-
-        throw &Errno::ENOENT unless $inode;
+        confess('No such file or directory') unless $inode;
 
         if ( $inode->link ) {
-            $hier = $hier->concat( $inode->readlink )
-              if $opts{'resolve_symlinks'} || $hier->count;
+            $hier = $hier->concat( $inode->readlink ) if $opts{'resolve_symlinks'} || $hier->count;
         }
         else {
             $dir = $inode;
@@ -212,7 +250,10 @@ directory when prefixed with a slash ('C</>').
 
 sub stat {
     my ( $self, $path ) = @_;
-    return $self->_find_inode( $path, 'resolve_symlinks' => 1 );
+    return $self->_find_inode(
+        $path,
+        'resolve_symlinks' => 1
+    );
 }
 
 =item C<$fs-E<gt>lstat($path)>
@@ -245,19 +286,16 @@ sub fstat {
 
 Change the current working directory to the path specified.  An
 C<$fs-E<gt>stat()> call will be used internally to lookup the inode for that
-path; an ENOTDIR will be thrown unless the inode found is a directory.  The
-internal current working directory pointer will be updated with the directory
-inode found; this same inode will also be returned.
+path; an exception "Not a directory" will be thrown unless the inode found is a
+directory.  The internal current working directory pointer will be updated with
+the directory inode found; this same inode will also be returned.
 
 =cut
 
 sub chdir {
     my ( $self, $path ) = @_;
     my $inode = $self->stat($path);
-
-    $! = 0;
-
-    throw &Errno::ENOTDIR unless $inode->dir;
+    confess('Not a directory') unless $inode->dir;
 
     return $self->{'cwd'} = $inode;
 }
@@ -267,17 +305,14 @@ sub chdir {
 When passed a file descriptor for a directory, update the internal pointer to
 the current working directory to that directory resolved from the file
 descriptor table, and return the same directory inode.  If the inode is not a
-directory, an ENOTDIR will be thrown.
+directory, an exception "Not a directory" will be thrown.
 
 =cut
 
 sub fchdir {
     my ( $self, $fd ) = @_;
     my $inode = $self->fstat($fd);
-
-    $! = 0;
-
-    throw &Errno::ENOTDIR unless $inode->dir;
+    confess('Not a directory') unless $inode->dir;
 
     return $self->{'cwd'} = $inode;
 }
@@ -327,7 +362,7 @@ affected inode will be returned.
 sub chmod {
     my ( $self, $path, $mode ) = @_;
     my $inode = $self->stat($path);
-
+    
     $inode->chmod($mode);
 
     return $inode;
@@ -344,7 +379,7 @@ to that inode will be returned.
 sub fchmod {
     my ( $self, $fd, $mode ) = @_;
     my $inode = $self->fstat($fd);
-
+    
     $inode->chmod($mode);
 
     return $inode;
@@ -356,9 +391,9 @@ sub fchmod {
 
 Create a new directory at the path specified, applying the permissions field in
 the mode value specified.  If no mode is specified, the default permissions of
-I<0777> will be modified by the current umask value.  An ENOTDIR exception will
-be thrown in case the intended parent of the directory to be created is not
-actually a directory itself.
+I<0777> will be modified by the current umask value.  A "Not a directory"
+exception will be thrown in case the intended parent of the directory to be
+created is not actually a directory itself.
 
 A reference to the newly-created directory inode will be returned.
 
@@ -387,24 +422,25 @@ the directory entry for the destination is empty.
 
 Links traversing filesystem mount points are not allowed.  This functionality
 is provided in the C<alias()> call provided by the L<Filesys::POSIX::Extensions>
-module.  Upon success, a reference to the inode for which a new link is to be
-created will be returned.
+module, which can be imported by C<$fs-E<gt>import_module()> at runtime.  Upon
+success, a reference to the inode for which a new link is to be created will be
+returned.
 
 Exceptions thrown:
 
 =over
 
-=item * EXDEV (Cross-device link)
+=item Cross-device link
 
 The inode resolved for the link source is not associated with the same device
 as the inode of the destination's parent directory.
 
-=item * EISDIR (Is a directory)
+=item Is a directory
 
 Thrown if the source inode is a directory.  Hard links can only be made for
 non-directory inodes.
 
-=item * EEXIST (File exists)
+=item File exists
 
 Thrown if an entry at the destination path already exists.
 
@@ -420,52 +456,47 @@ sub link {
     my $parent    = $self->stat( $hier->dirname );
     my $directory = $parent->directory;
 
-    $! = 0;
-
-    throw &Errno::EXDEV unless $inode->{'dev'} == $parent->{'dev'};
-    throw &Errno::EISDIR if $inode->dir;
-    throw &Errno::EEXIST if $directory->exists($name);
+    confess('Cross-device link') unless $inode->{'dev'} == $parent->{'dev'};
+    confess('Is a directory') if $inode->dir;
+    confess('File exists') if $directory->exists($name);
 
     $directory->set( $name, $inode );
 
     return $inode;
 }
 
-=item C<$fs-E<gt>symlink($old, $new)>
+=item C<$fs-E<gt>symlink($path, $dest)>
 
-The path in the first argument specified, C<$old>, is cleaned up using
-C<Filesys::POSIX::Path-E<gt>full>, and stored in a new symlink inode created
-in the location specified by C<$new>.  An EEXIST exception will be thrown if an
-inode at the path indicated by C<$new> exists.  A reference to the newly-created
-symlink inode will be returned.
+The path in the first argument specified, $path, is cleaned up using
+C<Filesys::POSIX::Path-E<gt>full()>, and stored in a new symlink inode created
+in the location specified by $dest.  An exception will be thrown if the
+destination exists.  A reference to the newly-created symlink inode will be
+returned.
 
 =cut
 
 sub symlink {
-    my ( $self, $old, $new ) = @_;
+    my ( $self, $path, $dest ) = @_;
     my $perms  = $S_IPERM ^ $self->{'umask'};
-    my $hier   = Filesys::POSIX::Path->new($new);
+    my $hier   = Filesys::POSIX::Path->new($dest);
     my $name   = $hier->basename;
     my $parent = $self->stat( $hier->dirname );
 
-    return $parent->child( $name, $S_IFLNK | $perms )->symlink( Filesys::POSIX::Path->full($old) );
+    return $parent->child( $name, $S_IFLNK | $perms )->symlink( Filesys::POSIX::Path->full($path) );
 }
 
 =item C<$fs-E<gt>readlink($path)>
 
 Using C<$fs-E<gt>lstat()> to resolve the given path for an inode, the symlink
-destination path associated with the inode is returned as a string.  An EINVAL
-exception is thrown unless the inode found is indeed a symlink.
+destination path associated with the inode is returned as a string.  A "Not a
+symlink" exception is thrown unless the inode found is indeed a symlink.
 
 =cut
 
 sub readlink {
     my ( $self, $path ) = @_;
     my $inode = $self->lstat($path);
-
-    $! = 0;
-
-    throw &Errno::EINVAL unless $inode->link;
+    confess('Not a symlink') unless $inode->link;
 
     return $inode->readlink;
 }
@@ -478,12 +509,12 @@ exceptions will be thrown in the event of certain errors:
 
 =over
 
-=item * ENOENT (No such file or directory)
+=item No such file or directory
 
 No entry was found in the path's parent directory for the item specified in the
 path.
 
-=item * EISDIR (Is a directory)
+=item Is a directory
 
 C<$fs-E<gt>unlink()> was called with a directory specified.
 C<$fs-E<gt>rmdir()> must be used instead for removing directory inodes.
@@ -503,10 +534,8 @@ sub unlink {
     my $directory = $parent->directory;
     my $inode     = $directory->get($name);
 
-    $! = 0;
-
-    throw &Errno::ENOENT unless $inode;
-    throw &Errno::EISDIR if $inode->dir;
+    confess('No such file or directory') unless $inode;
+    confess('Is a directory') if $inode->dir;
 
     $directory->delete($name);
 
@@ -515,11 +544,11 @@ sub unlink {
 
 =item C<$fs-E<gt>rename($old, $new)>
 
-Relocate the item specified by the C<$old> argument to the new path specified by
+Relocate the item specified by the $old argument to the new path specified by
 $new.
 
-Using C<$fs-E<gt>lstat>, the inode for the old pathname is resolved;
-C<$fs-E<gt>stat> is then used to resolve the path of the parent directory of
+Using $fs->lstat(), the inode for the old pathname is resolved;
+C<$fs-E<gt>stat()> is then used to resolve the path of the parent directory of
 the argument specified in C<$new>.
 
 If an inode exists at the path specified by C<$new>, it will be replaced by
@@ -538,27 +567,27 @@ The following exceptions are thrown for error conditions:
 
 =over
 
-=item * EPERM (Operation not permitted)
+=item Operation not permitted
 
 Currently, C<$fs-E<gt>rename()> cannot operate if the inode at the old location
 is an inode associated with a Filesys::POSIX::Real filesystem type.
 
-=item * EXDEV (Cross-device link)
+=item Cross-device link
 
 The inode at the old path does not exist on the same filesystem device as the
 inode of the parent directory specified in the new path.
 
-=item * ENOTDIR (Not a directory)
+=item Not a directory
 
 The old inode is a directory, but an existing inode found in the new path
 specified, is not.
 
-=item * EISDIR (Is a directory)
+=item Is a directory
 
 The old inode is not a directory, but an existing inode found in the new path
 specified, is.
 
-=item * ENOTEMPTY (Directory not empty)
+=item Directory not empty
 
 Both the old and new paths correspond to a directory, but the new path is not
 of an empty directory.
@@ -571,34 +600,29 @@ Upon success, a reference to the inode to be renamed will be returned.
 
 sub rename {
     my ( $self, $old, $new ) = @_;
+    my $hier      = Filesys::POSIX::Path->new($new);
+    my $name      = $hier->basename;
+    my $inode     = $self->lstat($old);
+    my $parent    = $self->stat( $hier->dirname );
+    my $directory = $parent->directory;
 
-    my $inode = $self->lstat($old);
+    confess('Operation not permitted') if ref $inode eq 'Filesys::POSIX::Real::Inode';
+    confess('Cross-device link') unless $inode->{'dev'} eq $parent->{'dev'};
 
-    my $old_hier   = Filesys::POSIX::Path->new($old);
-    my $old_name   = $old_hier->basename;
-    my $old_parent = $self->stat( $old_hier->dirname );
-    my $old_dir    = $old_parent->directory;
-
-    my $new_hier   = Filesys::POSIX::Path->new($new);
-    my $new_name   = $new_hier->basename;
-    my $new_parent = $self->stat( $new_hier->dirname );
-    my $new_dir    = $new_parent->directory;
-
-    $! = 0;
-
-    throw &Errno::EXDEV unless $inode->{'dev'} eq $new_parent->{'dev'};
-
-    if ( my $existing = $new_dir->get($new_name) ) {
+    if ( my $existing = $directory->get($name) ) {
         if ( $inode->dir ) {
-            throw &Errno::ENOTDIR   unless $existing->dir;
-            throw &Errno::ENOTEMPTY unless $existing->empty;
+            confess('Not a directory')     unless $existing->dir;
+            confess('Directory not empty') unless $existing->empty;
         }
         else {
-            throw &Errno::EISDIR if $existing->dir;
+            confess('Is a directory') if $existing->dir;
         }
     }
 
-    $new_dir->rename_member( $inode, $old_dir, $old_name, $new_name );
+    my $remove = $inode->dir ? 'rmdir' : 'unlink';
+    $self->$remove($old);
+
+    $directory->set( $name, $inode );
 
     return $inode;
 }
@@ -610,20 +634,20 @@ the following conditions:
 
 =over
 
-=item * ENOENT (No such file or directory)
+=item No such file or directory
 
 No inode exists by the name specified in the final component of the path in
 the parent directory specified in the path.
 
-=item * EBUSY (Device or resource busy)
+=item Device or resource busy
 
 The directory specified is an active mount point.
 
-=item * ENOTDIR (Not a directory)
+=item Not a directory
 
 The inode found at C<$path> is not a directory.
 
-=item * ENOTEMPTY (Directory not empty)
+=item Directory not empty
 
 The directory is not empty.
 
@@ -642,85 +666,15 @@ sub rmdir {
     my $directory = $parent->directory;
     my $inode     = $directory->get($name);
 
-    $! = 0;
-
-    throw &Errno::ENOENT unless $inode;
-
-    throw &Errno::EBUSY if $self->{'vfs'}->statfs(
-        $self->stat($path),
-        'exact' => 1, 'silent' => 1
-    );
-
-    throw &Errno::ENOTEMPTY unless $inode->empty;
+    confess('No such file or directory') unless $inode;
+    confess('Device or resource busy') if $self->{'vfs'}->statfs( $self->stat($path), 'exact' => 1, 'silent' => 1 );
+    confess('Directory not empty') unless $inode->empty;
 
     $directory->delete($name);
 
     return $inode;
 }
 
-=item C<$fs-E<gt>mknod($path, $mode)>
-
-=item C<$fs-E<gt>mknod($path, $mode, $dev)>
-
-Create a new inode at the specified C<$path>, with the inode permissions and
-format specified in the C<$mode> argument.  If C<$mode> specifies a C<$S_IFCHR>
-or C<$S_IFBLK> value, then the device number specified in C<$dev> will be given
-to the new inode.
-
-Code contained within the C<Filesys::POSIX> distribution assumes that the device
-identifier shall contain the major and minor numbers in separate 16-bit fields,
-in the following manner:
-
-    my $major = ($dev & 0xffff0000) >> 16;
-    my $minor =  $dev & 0x0000ffff;
-
-Returns a reference to a L<Filesys::POSIX::Inode> object upon success.
-
-=cut
-
-sub mknod {
-    my ( $self, $path, $mode, $dev ) = @_;
-    my $hier      = Filesys::POSIX::Path->new($path);
-    my $name      = $hier->basename;
-    my $parent    = $self->lstat( $hier->dirname );
-    my $directory = $parent->directory;
-
-    my $format = $mode & $S_IFMT;
-    my $perms  = $mode & $S_IPERM;
-
-    $! = 0;
-
-    throw &Errno::EINVAL unless $format;
-    throw &Errno::EEXIST if $directory->exists($name);
-
-    my $inode = $parent->child( $name, $format | $perms );
-
-    if ( $format == $S_IFCHR || $format == $S_IFBLK ) {
-        $inode->{'rdev'} = $dev;
-    }
-
-    return $inode;
-}
-
-=item C<$fs-E<gt>mkfifo($path, $mode)>
-
-Create a new FIFO device at the specified C<$path>, with the permissions listed
-in C<$mode>.  Internally, this function is a frontend to
-C<Filesys::POSIX-E<gt>mknod>.
-
-Returns a reference to a L<Filesys::POSIX::Inode> object upon success.
-
-=cut
-
-sub mkfifo {
-    my ( $self, $path, $mode ) = @_;
-
-    my $format = $S_IFIFO;
-    my $perms  = $mode & $S_IPERM;
-
-    return $self->mknod( $path, $format | $perms );
-}
-
 =back
 
 =cut
@@ -741,23 +695,6 @@ Filesys::POSIX.  For example, one method is provided that allows a developer to
 map a file or directory from the system's underlying, actual filesystem, into
 any arbitrary point in the virtual filesystem.
 
-=item L<Filesys::POSIX::Userland::Find>
-
-Provides the ability to perform breadth-first operations on file hierarchies
-within an instance of a C<Filesys::POSIX> filesystem, in a subset of the
-functionality provided in L<File::Find>.
-
-=item L<Filesys::POSIX::Userland::Tar>
-
-Provides an implementation of the POSIX ustar and certain aspects of the GNU tar
-standard.  Currently allows for the creation of tar archives based on
-hierarchies within a C<Filesys::POSIX> instance.
-
-=item L<Filesys::POSIX::Userland::Test>
-
-Provides a series of truth tests that can be performed on files and directories
-specified by paths.
-
 =back
 
 =head1 UTILITIES
@@ -783,11 +720,6 @@ Lists the requirements for writing modules that act as directory structures.
 
 Lists the requirements for writing modules that act as inodes.
 
-=item L<Filesys::POSIX::Module>
-
-Provides an interface for loading methods from modules that extend
-Filesys::POSIX.
-
 =back
 
 =head1 INTERNALS
@@ -823,22 +755,3 @@ Used by Filesys::POSIX, this module provides an implementation of a filesystem
 mount table and VFS inode resolution routines.
 
 =back
-
-=head1 AUTHOR
-
-Written by Xan Tronix <xan@cpan.org>
-
-=head1 CONTRIBUTORS
-
-=over
-
-=item Rikus Goodell <rikus.goodell@cpanel.net>
-
-=item Brian Carlson <brian.carlson@cpanel.net>
-
-=back
-
-=head1 COPYRIGHT
-
-Copyright (c) 2014, cPanel, Inc.  Distributed under the terms of the Perl
-Artistic license.
